/*
 * main-FESR-PiPhot.c
 *
 *  Created on: Apr 6, 2018
 *      Author: Vincent
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <complex.h>    			/* Standard Library of Complex Numbers */


#define COUCOU() printf("***I'm at line %d***\n",__LINE__);	// debug ;-)

// UNITS = GeV
#define HBC     0.19732				// h-bar c
#define MP      0.93827203			// proton mass
#define MPI     0.13803899			// do not confuse MPI (pion mass) and M_PI (math pi)
#define GP		13.54				// pion-nucleon couplings
#define EC		0.303				// electric charge
#define KAPPAP  1.78
#define KAPPAN  -1.91
#define ELABCUT 2.0					// limit of applicability of partial wave models
#define MAXBIN  500					// max number of point in the multipole files
#define LMAX    5					// max L for the multipoles
#if !defined( M_PI )
	#define M_PI 3.141592654		// otherwise it doesn't compile on FuturGrid
#endif


// Structure with all the kinematical quantities
struct Kin
{
	double m12;				// Mass particle 1
	double m22;				// Mass particle 2
	double m32;				// Mass particle 3
	double m42;				// Mass particle 4
	double complex s;				// Mandelstam s variable
    double complex t;				// Mandelstam t variable
    double complex phi;				// phi = stu + ...
    double complex ks;		// photon  momentum in s-channel frame
    double complex qs;		// pion    momentum in s-channel frame
    double complex kt;		// photon  momentum in t-channel frame
    double complex pt;		// nucleon momentum in t-channel frame
    double complex E2s;		// initial nucleon energy in s-channel frame
    double complex E4s;		// final   nucleon energy in s-channel frame
    double complex E2t;		// initial nucleon energy in t-channel frame
    double complex E4t;		// final   nucleon energy in t-channel frame
    double complex coss;	// cos \theta_s
    double complex sins;	// sin \theta_s
    double complex cossh;	// cos \theta_s/2
    double complex sinsh;	// sin \theta_s/2
    double complex cost;	// cos \theta_t
    double complex sint;	// sin \theta_t
    double complex costh;	// cos \theta_t/2
    double complex sinth;	// sin \theta_t/2
};



// Low Energy Functions
double fun_ImAi (double nu);
void fesr_ImAi (double t, int k[2], double cut, double fesr[13]);
void Smooth_Multipoles(double s, double wval[MAXBIN],
		double complex MltPoleIN[13][LMAX+1][MAXBIN], double complex MltPoleOUT[13][LMAX+1]);
void Read_Multipoles(char* filename, double complex MltPole[13][LMAX+1][MAXBIN], double wval[MAXBIN]);
void MltPole2F(double s, double t, double complex MltPole[5][LMAX+1], double complex CGLNF[5]);

void kinematics(double complex s, double complex t, double mass[5], struct Kin *var);
void CGLNA2F(struct Kin var, double complex Ai[5], double complex Fi[5]);
void CGLNF2A(struct Kin var, double complex Ai[5], double complex Fi[5]);
void amplHi(struct Kin var, double complex Fi[5], double complex Hi[5]);

// Functions
double Coef(double E, double mass2[5]);
double complex lambda(double complex a, double b, double c);
double Cos2T(double E, double z, double mass2[5]);
double T2Cos(double E, double t, double mass2[5]);
double int_adaptsimpson(double (*f)(double),double a,double b,double tau);
double simpson(double (*f)(double),double a,double b);


// need global variables for functions to integrate
int IA = 1;													// which Ai to integrate
int MOM = 0;												// moment of the FESR
double TGLOBAL = 0.0; 										// to pass t as a global variable
double WVAL[MAXBIN] = {0.0};								// W bin values of multipole model
double complex MLTPOLE[13][LMAX+1][MAXBIN] = {{{0}}};	// Multipoles all isospin combinations


int main(void) {
	clock_t begin = clock();

	/* READ SIMULATION PARAMETERS. The format is
	 * Elab tmin tmax dt cutoff k1 k2 wmin wmax dw t0
	 * Elab				Lab energy for printing the observables for high energy
	 * tmin tmax dt		interval in t for the FESR
	 * wmin wmax dw		interval in w for printing Ai
	 * t0				print Ai at t = t0
	 * cutoff			W max for the FESR
	 * k1 k2			moments of the FESR
	 */
	double parSimu[20] = {0};
	FILE *fileParSimu = fopen("Simulation-param.txt","r");
	for (int i = 0; i<20; i++){ fscanf(fileParSimu, "%lf", &parSimu[i]); }
	fclose(fileParSimu);

	//double Elab = parSimu[0]; 	// not needed in this code
	double t, tdeb = parSimu[1], tfin = parSimu[2], dt = parSimu[3], cut = parSimu[4];
	double s, wdeb = parSimu[7], wfin = parSimu[8], dw = parSimu[9], t0  = parSimu[10];


	// k[0] needs to be odd ; k[1] needs to be even
	int k1 = parSimu[5], k2 = parSimu[6], k[2];
	k[0] = floor((k1-1)/2.0)*2+1;
	k[1] = floor(k2/2.0)*2;


	// LOAD MULTIPOLES
	int ind;
	char name[40];
	struct Kin var;
	double complex Ai[5] = {0}, Fi[5] = {0};
	double complex MltPoleW[5][LMAX+1] = {{0}};
	double complex MltPoleAllW[13][LMAX+1] = {{0}};
	double mass2[5] = {0, 0, MP*MP, MPI*MPI, MP*MP};

	FILE *fileName = fopen("multipoles-model.txt","r");
	fscanf(fileName, "%s", name); fclose(fileName);

	Read_Multipoles(name, MLTPOLE, WVAL);

	// FESR FROM MULTIPOLES
	double fesr[13] = {0}, err[13] = {0};
	FILE *fileFESR_Low_All = fopen("FESR-Low-All.txt","w");
	FILE *fileFESR_Low_err = fopen("FESR-Low-All-err.txt","w");
	for(t = tdeb ; t > tfin  ; t += dt){

		fesr_ImAi(t, k, cut, fesr);
		// print FESR in file
		fprintf(fileFESR_Low_All,	"%8.4f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f\n", t,
					fesr[1], fesr[2],  fesr[3],
					fesr[1] + t*fesr[4], fesr[2] + t*fesr[5], fesr[3] + t*fesr[6],
					fesr[7], fesr[8], fesr[9],
					fesr[10], fesr[11], fesr[12]);
		// the error is just the maximum of each fesr
		for(int iA = 1; iA<13; iA++){
			if(fabs(fesr[iA]) > err[iA]){ err[iA] = fabs(fesr[iA]);}
		}
		err[4] = 0;
		if( fabs(fesr[1] + t*fesr[4]) > err[4] ){ err[4] = fabs(fesr[1] + t*fesr[4]);}
		err[5] = 0;
		if( fabs(fesr[2] + t*fesr[5]) > err[5] ){ err[5] = fabs(fesr[2] + t*fesr[5]);}
		err[6] = 0;
		if( fabs(fesr[3] + t*fesr[6]) > err[6] ){ err[6] = fabs(fesr[3] + t*fesr[6]);}
	}

	// print the error, 20% of the max value, in file
	double fac = 0.2;
	for(t = tdeb ; t > tfin  ; t += dt){
		fprintf(fileFESR_Low_err,	"%8.4f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f\n", t,
					fac*err[1], fac*err[2], fac*err[3],
					fac*err[4], fac*err[5], fac*err[6],
					fac*err[7], fac*err[8], fac*err[9],
					fac*err[10], fac*err[11], fac*err[12]);
	}
	fclose(fileFESR_Low_All);fclose(fileFESR_Low_err);

	// PRINT Ai at t = t0
	FILE *fileAi0_Low_T0 = fopen("Ai0-Low-T0.txt","w");
	FILE *fileAip_Low_T0 = fopen("Aip-Low-T0.txt","w");
	FILE *fileAim_Low_T0 = fopen("Aim-Low-T0.txt","w");


	for(double w = wdeb ; w < wfin  ; w += dw){

		s = w*w;
		kinematics(s, t0, mass2, &var);
		Smooth_Multipoles(s, WVAL, MLTPOLE, MltPoleAllW);

		ind = 0;
		// select if isospin is (0,+,-) = (0,1,2) with ind = mod(IA-1,3)
		for(int iL=0; iL<= LMAX; iL++){
		for(int i =1; i <= 4   ;  i++){
			MltPoleW[i][iL] = MltPoleAllW[4*ind+i][iL];
		}}

		MltPole2F(s, t0, MltPoleW, Fi);				// return Fi[5]
		CGLNF2A(var, Ai, Fi);						// return Ai[5]

		fprintf(fileAi0_Low_T0,"%5.3f %f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f \n", sqrt(s),
				creal(Ai[1]), cimag(Ai[1]), creal(Ai[2]+t0*Ai[2]), cimag(Ai[2]+t0*Ai[2]),
				creal(Ai[3]), cimag(Ai[3]), creal(Ai[4]), cimag(Ai[4]));

		ind = 1;
		for(int iL=0; iL<= LMAX; iL++){
		for(int i =1; i <= 4   ;  i++){
			MltPoleW[i][iL] = MltPoleAllW[4*ind+i][iL];
		}}

		MltPole2F(s, t0, MltPoleW, Fi);					// return Fi[5]
		CGLNF2A(var, Ai, Fi);							// return Ai[5]
		fprintf(fileAip_Low_T0,"%5.3f %f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f \n", sqrt(s),
				creal(Ai[1]), cimag(Ai[1]), creal(Ai[2]+t0*Ai[2]), cimag(Ai[2]+t0*Ai[2]),
				creal(Ai[3]), cimag(Ai[3]), creal(Ai[4]), cimag(Ai[4]));


		ind = 2;
		for(int iL=0; iL<= LMAX; iL++){
		for(int i =1; i <= 4   ;  i++){
			MltPoleW[i][iL] = MltPoleAllW[4*ind+i][iL];
		}}


		MltPole2F(s, t0, MltPoleW, Fi);					// return Fi[5]
		CGLNF2A(var, Ai, Fi);							// return Ai[5]
		fprintf(fileAim_Low_T0,"%5.3f %f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f \n", sqrt(s),
				creal(Ai[1]), cimag(Ai[1]), creal(Ai[2]+t0*Ai[2]), cimag(Ai[2]+t0*Ai[2]),
				creal(Ai[3]), cimag(Ai[3]), creal(Ai[4]), cimag(Ai[4]));


	}
	fclose(fileAi0_Low_T0);fclose(fileAip_Low_T0);fclose(fileAim_Low_T0);


	clock_t end = clock();
	double time_spent = (double)(end - begin) / CLOCKS_PER_SEC;
	printf("\n");
	printf("time spent: %f s", time_spent);


}

// *********************************************************************************
void fesr_ImAi(double t, int k[2], double cut, double fesr[13]){
	/*
	 * return FESR for A1,...,A4 for the multipoles uploaded
	 * cut is Wmax in GeV
	 * k[0] is odd  --> A1,A2,A4
	 * k[1] is even --> A3
	 */
	TGLOBAL = t;

	double nuFin = (cut*cut - MP*MP)/(2*MP) + (t-MPI*MPI)/(4*MP);
	double nuDeb = MPI + (t + MPI*MPI)/(4*MP) + 0.005;
	double dok = 0.00001; 					// precision
	double nuB = (t - MPI*MPI)/(4*MP);
	double fac = M_PI*GP*EC/(2*MP);
	double sum = (KAPPAP+KAPPAN)/(4*MP), diff = (KAPPAP-KAPPAN)/(4*MP);
	double born[13] = {0, -fac/2.0, -fac/2.0, -fac/2.0,
						fac/(t-MPI*MPI), fac/(t-MPI*MPI), fac/(t-MPI*MPI),
						fac*sum, fac*diff, fac*diff,
						fac*sum, fac*diff, fac*diff};

	// k[0] is odd ; k[1] is even
	// iA = {1,2,4,5,9,10,11} need k odd
	// iA = {3,6,7,8,12}      need k even
	int mo[13]  = {0, 0,0,1,0,0,1,1,1,0,0,0,1};
	double integral[13] = {0};
	for(int i = 1; i <=12 ; i++){
		MOM = k[mo[i]];
		IA = i; integral[IA] = int_adaptsimpson (*fun_ImAi, nuDeb, nuFin, dok);
		// add Borm term
		fesr[IA] = ( born[IA]*pow(nuB, MOM) + integral[IA])/pow(nuFin, MOM + 1);
	}
	return;
}


//*************************************************************************
double fun_ImAi (double nu){
	struct Kin var;
	double s = MP*MP + 2*MP*nu + (MPI*MPI-TGLOBAL)/2;
	int ind = (IA-1) % 3;
	double mass2[5] = {0, 0, MP*MP, MPI*MPI, MP*MP};
	double complex Ai[5] = {0}, Fi[5] = {0};
	double complex MltPoleW[5][LMAX+1] = {{0}};
	double complex MltPoleAllW[13][LMAX+1] = {{0}};

	kinematics(s, TGLOBAL, mass2, &var);
	Smooth_Multipoles(s, WVAL, MLTPOLE, MltPoleAllW);	// compute all 12 multipoles for s
	// select if isospin is (0,+,-) = (0,1,2) with ind = mod(IA-1,3)
	for(int iL=0; iL<= LMAX; iL++){
	for(int i =1; i <= 4   ;  i++){
		MltPoleW[i][iL] = MltPoleAllW[4*ind+i][iL];
	}}

	MltPole2F(s, TGLOBAL, MltPoleW, Fi);				// return Fi[5]
	CGLNF2A(var, Ai, Fi);								// return Ai[5]

	return pow(nu,MOM)*cimag(Ai[(IA-1)/3+1]);
}

// *********************************************************************************
void Smooth_Multipoles(double s, double wval[MAXBIN],
		double complex MltPoleIN[13][LMAX+1][MAXBIN],
		double complex MltPoleOUT[13][LMAX+1]){
	/*
	 * retun multipoles interpolated between wbins
	 * input s is in GeV^2 but wval is in MeV
	 */
	double w = sqrt(s);
	double binSize = (wval[1]-wval[0])/1000;
	double wdebut  = wval[0]/1000;
	int ind = floor((w - wdebut)/binSize + 1);
	double eps = (w-wval[ind]/1000)/binSize;

	// Linear extrapolation between W bins
	for(int L = 0 ; L <= LMAX ; L++){
	for(int i = 1; i<13 ;i++){
		MltPoleOUT[i][L] = (1-eps)*MltPoleIN[i][L][ind] + eps*MltPoleIN[i][L][ind+1];
	}}
	return;
}


// **************************************************************************************************
void Read_Multipoles(char* filename, double complex MltPole[13][LMAX+1][MAXBIN], double wval[MAXBIN]){
	/*
	 * Read multipoles from files.
	 * Output wval is in MeV
	 * return double complex MltPole[13][10]
	 * MltPole[1-4][L]  is E_{L+}^(0), E_{L-}^(0), M_{L+}^(0), M_{L-}^(0)
	 * MltPole[5-8][L]  is E_{L+}^(+), E_{L-}^(+), M_{L+}^(+), M_{L-}^(+)
	 * MltPole[9-12][L] is E_{L+}^(-), E_{L-}^(-), M_{L+}^(-), M_{L-}^(-)
	 *
	 * the input files format is (for each L)
	 * W(MeV) 	ReE_{L+}^(0)  ImE_{L+}^(0)  ReE_{L-}^(0) ImE_{L-}^(0)  ReM_{L+}^(0) ImM_{L+}^(0) ReM_{L-}^(0) ImM_{L-}^(0)
	 * 			ReE_{L+}^(+)  ImE_{L+}^(+)  ReE_{L-}^(+) ImE_{L-}^(+)  ReM_{L+}^(+) ImM_{L+}^(+) ReM_{L-}^(+) ImM_{L-}^(+)
	 * 			ReE_{L+}^(-)  ImE_{L+}^(-)  ReE_{L-}^(-) ImE_{L-}^(-)  ReM_{L+}^(-) ImM_{L+}^(-) ReM_{L-}^(-) ImM_{L-}^(-)
	 * there is then 25 numbers on each line.
	 * MAXIMUM L is LMAX
	 * The multipoles are in mFm
	 */

	// Need to read Re and Im separately from file
	double ReMlt[13][LMAX+1][MAXBIN] = {{{0}}};
	double ImMlt[13][LMAX+1][MAXBIN] = {{{0}}};

	double wvaltmp[MAXBIN] = {0};
	char namefull[100], Ltmp[10];
	//char path[100] = "/Users/Vincent/Dropbox/Workspace/FESR-PiPhot/Multipoles/";
	char path[100] = "/var/local/sites/cgl.soic.indiana.edu/jpac/FESR-PiPhot/multipoles/";
	strcat(path,filename);

	// loop over each L ; one file per L
	for(int L=0 ; L<=LMAX ; L++){
		strcpy(namefull,  path);
		sprintf(Ltmp, "%i", L);
		strcat(strcat(strcat(namefull, "-L"), Ltmp),".dat");

		FILE *file = fopen(namefull,"r");
		// exit the loop if no file found
		if (file == NULL) { fclose(file); printf("no file L = %i\n", L);  break ;	}

		for(int i = 0 ; i <= MAXBIN ; i++){
			fscanf(file,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf \n",
					&wvaltmp[i],
					&ReMlt[1][L][i], &ImMlt[1][L][i], &ReMlt[2][L][i], &ImMlt[2][L][i],
					&ReMlt[3][L][i], &ImMlt[3][L][i], &ReMlt[4][L][i], &ImMlt[4][L][i],
					&ReMlt[5][L][i], &ImMlt[5][L][i], &ReMlt[6][L][i], &ImMlt[6][L][i],
					&ReMlt[7][L][i], &ImMlt[7][L][i], &ReMlt[8][L][i], &ImMlt[8][L][i],
					&ReMlt[9][L][i], &ImMlt[9][L][i], &ReMlt[10][L][i], &ImMlt[10][L][i],
					&ReMlt[11][L][i], &ImMlt[11][L][i], &ReMlt[12][L][i], &ImMlt[12][L][i]);

			for(int j = 1; j <13 ; j++){
				MltPole[j][L][i] = ReMlt[j][L][i] + I*ImMlt[j][L][i];
			}

			// take the W values from the L = 0 file
			if(L==0){wval[i] = wvaltmp[i];}
		}
		fclose(file);
	}
	return;
}


// *********************************************************************************************
void MltPole2F(double s, double t, double complex MltPole[5][LMAX+1], double complex CGLNF[5]){
	/* input:
	 * MltPole[1][L] is E_{L+}
	 * MltPole[2][L] is E_{L-}
	 * MltPole[3][L] is M_{L+}
	 * MltPole[4][L] is M_{L-}
	 *
	 * input multipoles are in mFm
	 * outpu CGLN Fi in mFm
	 * input s,t are in GeV^2
	 */
	int k;

	// kinematics
	double w = sqrt(s);
	double som = 2*MP*MP + MPI*MPI;
	double u = -s - t + som;
	double kcm = (s - MP*MP)/(2.0*w);
	double Ex = (s + MP*MP - MPI*MPI)/(2*w);
	double qcm = sqrt(Ex*Ex - MP*MP);
	double z = ( s *(t-u) - MP*MP*(MPI*MPI - MP*MP))/(4*s*kcm*qcm);	// z = cos theta
	//double tmin = pow(MPI,4)/4/s - pow(qcm - kcm,2);
	//double tmax = tmin - 4*kcm*qcm;

	double pol1p[10] = {0}; // First derivative of Legendre polynomials
	double pol2p[10] = {0}; // second derivative of Legendre polynomials
	pol1p[1] = 1.0;
	pol1p[2] = 3*z;
	pol1p[3] = (-3 + 15*z*z)/2.0;
	pol1p[4] = (-60*z + 140*z*z*z)/8;
	pol1p[5] = (15 - 210*z*z + 315*z*z*z*z)/8;
	pol1p[6] = (210*z - 1260*z*z*z + 1386*pow(z,5))/16;
	pol1p[7] = (-35 + 945*z*z - 3465*pow(z,4) + 3003*pow(z,6))/16;
	pol1p[8] = (-2520*z + 27720*z*z*z - 72072*pow(z,5) + 51480*pow(z,7))/128;
	pol1p[9] = (315 - 13860*z*z + 90090*pow(z,4) - 180180*pow(z,6) + 109395*pow(z,8))/128;

	pol2p[2] = 3;
	pol2p[3] = 15*z;
	pol2p[4] = (-60+420*z*z)/8;
	pol2p[5] = (-420*z + 1260*z*z*z)/8.0;
	pol2p[6] = (210 - 3780*z*z + 6930*pow(z,4))/16.0;
	pol2p[7] = (1890*z - 13860*z*z*z + 18018*pow(z,5))/16.0;
	pol2p[8] = (-2520 + 83160*z*z - 360360*pow(z,4) + 360360*pow(z,6) )/128;
	pol2p[9] = ((-27720*z + 360360*z*z*z - 1081080*pow(z,5) + 875160*pow(z,7)))/128;

	CGLNF[1] = 0; CGLNF[2] = 0; CGLNF[3] = 0; CGLNF[4] = 0;
	for( int L=0; L<=LMAX; L++){
		k = L-1; if(L==0){k = 0;}
		/*
		* MltPole[1][L] is E_{L+}
		* MltPole[2][L] is E_{L-}
	    * MltPole[3][L] is M_{L+}
		* MltPole[4][L] is M_{L-}
		*/
		CGLNF[1] = CGLNF[1] + (L*MltPole[3][L] + MltPole[1][L])*pol1p[L+1];
		CGLNF[1] = CGLNF[1] + ( (L+1)*MltPole[4][L] + MltPole[2][L] )*pol1p[k];
		CGLNF[2] = CGLNF[2] + ( (L+1)*MltPole[3][L] +L*MltPole[4][L])*pol1p[L];
		CGLNF[3] = CGLNF[3] + (MltPole[1][L] - MltPole[3][L])*pol2p[L+1];
		CGLNF[3] = CGLNF[3] + (MltPole[2][L] + MltPole[4][L])*pol2p[L-1];
		CGLNF[4] = CGLNF[4] + (MltPole[3][L] - MltPole[1][L] - MltPole[4][L] - MltPole[2][L])*pol2p[L];
	}
	return;
}



// *********************************************************************************
void kinematics(double complex s, double complex t, double mass2[5], struct Kin *var)
{
	double m12, m22, m32, m42; 	// masses squared
	double complex t0, t1, u ;

	var->s = s;
	var->t = t;

	var->m12 = mass2[1];
	var->m22 = mass2[2];
	var->m32 = mass2[3];
	var->m42 = mass2[4];

	m12 = mass2[1] ;
	m22 = mass2[2] ;
	m32 = mass2[3] ;
	m42 = mass2[4] ;

	var->ks = csqrt( lambda(s, m12, m22) / 4 / s );
	var->qs = csqrt( lambda(s, m32, m42) / 4 / s );
	var->kt = csqrt( lambda(t, m12, m32) / 4 / t );
	var->pt = csqrt( lambda(t, m22, m42) / 4 / t );

	// nuclei energies in s- and t-channel frames
	var->E2s = ( s + m22 - m12 ) / 2 / csqrt(s);
	var->E4s = ( s + m42 - m32 ) / 2 / csqrt(s);
	var->E2t = ( t + m22 - m42 ) / 2 / csqrt(t);
	var->E4t = ( t + m42 - m22 ) / 2 / csqrt(t);

	// t1 = tmin ; t0 = tmax ; t0 < t < t1
	t1 = cpow(m12 - m32 - m22 + m42,2)/(4*s) - (var->ks - var->qs) * (var->ks - var->qs);
	t0 = t1 - 4 * var->ks *var->qs ;
	u = - s - t + m12 + m22 + m32 + m42 ;	// Mandelstam s variable
	var->phi = s * (t - t1) * (t0 - t) ;	// Kibble function

	var->coss = 1 + (t - t1)/(2 * var->qs * var->ks) ;
	//var->cost = (t*(s-u) + (m12-m32) * (m22-m42) )/(4 * t * var->pt * var->kt) ;
	var->cost = (t*(s-u) + (m12-m32) * (m22-m42) )/(csqrt(lambda(var->t,m12,m32) * lambda(var->t,m22,m42) ) ) ;
	var->sins = csqrt( var->phi / s ) / ( 2 * var->qs * var->ks) ;
	var->sint = csqrt( var->phi / t ) / ( 2 * var->kt * var->pt) ;
	var->cossh = csqrt( (1 + var->coss ) / 2 );
	var->sinsh = csqrt( (1 - var->coss ) / 2 );
	var->costh = csqrt( (1 + var->cost ) / 2 );
	var->sinth = csqrt( (1 - var->cost ) / 2 );

	return ;
}

// *********************************************************************************
void CGLNF2A(struct Kin var, double complex Ai[5], double complex Fi[5]){
	/*
	 * Compute CGLN Ai(s,t) from CGLN Fi(s,t)
	 * Ai are in GeV^{-2,-4,-3,-3}
	 * Fi are in mFm
	 */
	double complex w;
	double complex m2 = sqrt(var.m22);
	double complex m4 = sqrt(var.m42);
	double complex E2p, E2m, E4p, E4m;
	double complex Fit[5] = {0};

	w = csqrt(var.s);
	E2p = var.E2s + m2 ;
	E2m = var.E2s - m2 ;
	E4p = var.E4s + m4 ;
	E4m = var.E4s - m4 ;

	double complex q = sqrt(E4p*E4m);
	double complex kq = (var.t-var.m32)/(2.0*q);
	double complex tmp = (var.t - var.m32 + 2*var.s - 2*var.m22)/(2*q*(sqrt(var.m22) + w));

	// kinematical factors
	Fit[1] = Fi[1]/sqrt(E2p*E4p);
	Fit[2] = Fi[2]/sqrt(E2m*E4m);
	Fit[3] = Fi[3]/sqrt(E2m*E4p);
	Fit[4] = Fi[4]/sqrt(E2p*E4m);

	Ai[1] =  (w + m4)*Fit[1] + (m4 - w)*(w - m2)/(w + m2)*Fit[2]
			+ (m2 + m4)/(m2 + w)*kq*(Fit[3] + Fit[4]);

	Ai[2] = (w - m2)/(w + m2)/q*Fit[3] - 1/q*Fit[4];
	Ai[3] = Fit[1] + (w - m2)/(w + m2)*Fit[2] + tmp*(Fit[3] + Fit[4]);
	Ai[4] = Fit[1] + (w - m2)/(w + m2)*Fit[2] + kq/(m2 + w)*(Fit[3] + Fit[4]);

	// conversion in GeV^-1
	for(int i = 1 ; i<=4 ; i++){
		Ai[i] = Ai[i] * 8.0*M_PI*w/HBC;
		Ai[i] = Ai[i] / (2*var.s - 2*w*m2) ;
		Ai[i] = Ai[i] / 1000.0;
	}
	return ;
}

// *********************************************************************************
void CGLNA2F(struct Kin var, double complex Ai[5], double complex Fi[5]){
	/*
	 * Compute CGLN Fi(s,t) from CGLN Ai(s,t)
	 * Ai are in GeV^-{2,4,3,3}
	 * Fi are in mFm
	 */
	double complex w;
	double complex m2 = sqrt(var.m22);
	double complex m4 = sqrt(var.m42);
	double complex E2p, E2m, E4p, E4m;

	w = csqrt(var.s);
	E2p = var.E2s + m2 ;
	E2m = var.E2s - m2 ;
	E4p = var.E4s + m4 ;
	E4m = var.E4s - m4 ;

	Fi[1] = (w - m2) * Ai[1] + (var.m32 - var.t)/2 * (Ai[3] - Ai[4]);
	Fi[1] = Fi[1]  + ( w - m2 ) * ( w - m4 ) * Ai[4];
	Fi[1] = Fi[1] * csqrt( E2p * E4p ) / ( 8 * M_PI * w);

	Fi[2] = -(w + m2) * Ai[1] + (var.m32 - var.t)/2 * (Ai[3] - Ai[4]);
	Fi[2] = Fi[2]  + ( w + m2 ) * ( w + m4 ) * Ai[4];
	Fi[2] = Fi[2] * csqrt( E2m * E4m ) / ( 8 * M_PI * w);

	Fi[3] = (w + m2) * (  (w - m2)*Ai[2] + Ai[3] - Ai[4] );
	Fi[3] = Fi[3] * csqrt( E2m * E4p) * var.qs  / ( 8 * M_PI * w);

	Fi[4] = (w - m2) * ( -(w + m2)*Ai[2] + Ai[3] - Ai[4] );
	Fi[4] = Fi[4] * csqrt( E2p * E4m) * var.qs  / ( 8 * M_PI * w);

	// above Fi[] are in GeV^-1
	// conversion in mFm
	// HBC = 0.19732 fm.GeV ; 1b = 100 fm^2 ; HBC^2 = (0.19732)^2 (10 mb).GeV^2
	// 10^4 HBC^2 = 389.35 mub.GeV^2
	double hbarc = 197.32; // 1000*hbar*c
	Fi[1] = hbarc*Fi[1];
	Fi[2] = hbarc*Fi[2];
	Fi[3] = hbarc*Fi[3];
	Fi[4] = hbarc*Fi[4];

	return;
}

// *********************************************************************************
void amplHi(struct Kin var, double complex Fi[5], double complex Hi[5]){

	Hi[1] = -1/sqrt(2) * var.sins * var.cossh * (Fi[3] + Fi[4]);
	Hi[3] =  1/sqrt(2) * var.sins * var.sinsh * (Fi[3] - Fi[4]);
	Hi[2] = sqrt(2) * var.cossh * (Fi[2] - Fi[1] )  + Hi[3];
	Hi[4] = sqrt(2) * var.sinsh * (Fi[2] + Fi[1] )  - Hi[1];

	return;
}

// ----------------------- FUNCTIONS  ---------------------------------

// ********************************************************************
double Cos2T(double E, double z, double mass2[5]){
	/*
	 * Compute t from Ecm = W and cos = z
	 */
	double t;
	double s = E*E;
	double ma2, mb2, mc2, md2;			// squared masses of the particles
	ma2 = mass2[1]; mb2 = mass2[2]; mc2 = mass2[3]; md2 = mass2[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma2, mb2) /4/s);
	qf = sqrt(lambda(s, mc2, md2) /4/s);
	double som = ma2 + mb2+ mc2 + md2;

	t = 2*qi*qf*z + som/2 - s/2 - (ma2 - mb2) * (mc2 - md2) /2/s ;
	return t;
}

// ********************************************************************
double T2Cos(double E, double t, double mass2[5]){
	/*
	 * Compute z = cos from Ecm = W
	 */
	double z;
	double s = E*E;
	double ma2, mb2, mc2, md2;			// squared masses of the particles
	ma2 = mass2[1]; mb2 = mass2[2]; mc2 = mass2[3]; md2 = mass2[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma2, mb2) /4/s);
	qf = sqrt(lambda(s, mc2, md2) /4/s);
	double som = ma2 + mb2 + mc2 + md2;

	z = (t + (ma2 - mb2)*(mc2 - md2)/(2.0*s) - som/2.0 + s/2.0 ) /(2*qi*qf);
	return z;
}

// ********************************************************************
double Coef(double E, double mass2[5]){
	/*
	 * dSig/dt = Coef dSig/dOmega
	 * Coef = Pi/qi/qf in GeV^-2
	 */
	double s = E*E;
	double ma2, mb2, mc2, md2;			// squared masses of the particles
	ma2 = mass2[1]; mb2 = mass2[2]; mc2 = mass2[3]; md2 = mass2[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma2, mb2) /4/s);
	qf = sqrt(lambda(s, mc2, md2) /4/s);
	return M_PI/qi/qf;
}

// ********************************************************************
double complex lambda(double complex a, double b, double c){
// triangle function
	return a*a + b*b + c*c - 2*(a*b + b*c + c*a);
}

// ********************************************************************
double simpson(double (*f)(double),double a,double b){
	return (b-a)/6*(f(a)+4*f((a+b)/2)+f(b));
}

// ********************************************************************
double int_adaptsimpson(double (*f)(double),double a,double b,double tau){
	double m=(a+b)/2;
	double Sab=simpson(f,a,b),Sam=simpson(f,a,m),Smb=simpson(f,m,b);
	if (fabs(Sab-Sam-Smb)/15<tau)
	  return Sam+Smb+(Sam+Smb-Sab)/15;
	else
	  return int_adaptsimpson(f,a,m,tau)+int_adaptsimpson(f,m,b,tau);
}

